import UIKit

var str = "Hello, playground"

// MARK: - Node

public class Node<T> {
    typealias NodeType = Node<T>                            // just to shorten definitions
    var value: T                                            // that's the object we want to keep in the list
    var next: NodeType?                                     // since this is a class, that's the pointer to the next node
    weak var previous: NodeType?                            // and thats the one the previous node
    var nextOrFirst: NodeType? {                            // a computed variable to make endless loops possible
        if next == nil {                                    // we reached the end of the list
            if let l = list {                               // just to be safe
                return l.head                               // start again from the top
            } else {                                        // must have unlinked the object
                return nil                                  // can't do anything with it
            }
        } else {                                            // not reached end yet
            return next                                     // take the next node
        }
    }

    fileprivate weak var list: LinkedList<T>?               // let's keep a pointer to the list as well

    init(value: T, list: LinkedList<T>) {                   // let's initialize our node
        self.value = value                                  // that's the object we want to keep
        self.list = list                                    // and remember our list as well
    }

    public func removeFromList() -> T {                     // let's remove this node from the list
        if let l = list {                                   // yes, we do know our list
            return l.remove(node: self)                     // let the list do it's job
        } else {                                            // oh no, no list known (any more)
            return value                                    // just return our object
        }
    }
}

// MARK: - Iterator

public struct LinkedListIterator<T>: IteratorProtocol {     // let's create an iterator for our linked list
    public typealias Element = Node<T>                      // the type of each node in the list
    
    private var currentNode: Element?                       // the current node in the iteration
    
    fileprivate init(startNode: Element?) {                 // let's initiate the iterator
        currentNode = startNode                             // just set the current node
    }
    
    public mutating func next() -> LinkedListIterator.Element? {    // get the next element in the list
        let node = currentNode                              // take the current node
        currentNode = currentNode?.next                     // and remember the next one, so the current node may
                                                            // be invalidated in the iteration
        return node                                         // return the node (not only the object, i.e. node?.value)
    }
}

// MARK: - LinkedList

enum CodingError: Error {                                   // enum for coding errors
    case decoding(String)
}

final public class LinkedList<T>: NSObject, NSCoding, Sequence {    // the linked list, finally, conforming to the sequence and the codable protocol
    
    public typealias Iterator = LinkedListIterator<T>       // the type of our iterator
    
    public func makeIterator() -> LinkedList.Iterator {     // let's create our iterator
        return LinkedListIterator(startNode: head)          // initialise it with the head of our list
    }
    
    typealias NodeType = Node<T>                            // the type of our nodes
    fileprivate var head: NodeType?                         // the first element of our list
    private var tail: NodeType?                             // and the last one; both may be nil
    
    public fileprivate(set) var count: Int = 0              // the number of elements in the list at any given time
    
    override public var description: String {
        var text = "["
        var node = head
        while node != nil {
            text += "\(node!.value)"
            node = node!.next
            if node != nil { text += ", " }
        }
        return text + "]"
    }
    
    override init() {                                       // just needed to add further inits
        super.init()                                        // let super do its job
    }

    public init<S: Sequence>(_ elements: S) where S.Iterator.Element == T {     // create the list from an other sequence
        
        super.init()
        for element in elements {                           // iterate over all elements in the supplied sequence
            self.append(value: element)                     // and extend our list
        }
    }

    required public init?(coder aDecoder: NSCoder) {        // create the list from an archive (NSCoding)
        
        super.init()                                        // let super do its job
        if T.self is PropertyListReadable.Type {            // the objects are PropertyListReadable (i.e. structs or enums)
            if let list = aDecoder.decodeObject() as? [PropertyList] {  // lets decode the list as an array of ProperList
                for object in list {                        // and iterate over all elements
                    if let item = (T.self as? PropertyListReadable.Type)?.init(propertyListRepresentation: object) {
                        self.append(value: item as! T)      // initialize the struct or enum and append it to the list
                    }
                }
            }
        } else if T.self is NSCoding.Type {                 // instead its a NSCoding compliant class
            if let list = aDecoder.decodeObject() as? [T] { // let it decode directly to an array of objects
                for object in list {                        // iterate over all elements
                    self.append(value: object)              // and append it to our list
                }
            }
        }
    }
    
    public func encode(with aCoder: NSCoder) {              // archive our list
        
        var next = head                                     // start at the top
        if T.self is PropertyListReadable.Type {            // our objects are PropertyListReadable (i.e. structs or enums)
            var list = [PropertyList]()                     // create an array of PropertyLists
            while next != nil {                             // and iterate over the whole list
                if let item = next!.value as? PropertyListReadable {    // each item is PropertyListReadable
                    list.append(item.propertyListRepresentation())      // append the propertyListRepresentation
                }
                next = next!.next                           // and get the next item
            }
            aCoder.encode(list)                             // finally let the Encoder encode the resulting array
        } else if T.self is NSCoding.Type {                 // instead its a NSCoding compliant class
            var list = [T]()                                // create an array of the objects
            while next != nil {                             // iterate over the whole list
                list.append(next!.value)                    // and append our objects
                next = next!.next                           // get the next item
            }
            aCoder.encode(list)                             // finally let the Encoder encode the resulting array
        }
    }

    public var isEmpty: Bool {                              // convenience property to test for emptiness
        return head == nil                                  // if the head is nil the list is empty
    }
    
    public var first: Node<T>? {                            // get the head of the list
        return head
    }
    
    public var last: Node<T>? {                             // and its tail
        return tail
    }
    
    public func append(value: T) {                          // append a new node at the end of the list
        let newNode = Node(value: value, list: self)        // create the node with the object
        if let tailNode = tail {                            // we do have an end (list is not empty)
            newNode.previous = tailNode                     // remember that node in our new one
            tailNode.next = newNode                         // and make the new node the new tail
        } else {                                            // the list is empty
            head = newNode                                  // make our new node the head of the list
        }
        tail = newNode                                      // our new node is the (new) end of the list
        count += 1                                          // and increase our counter
    }
    
    public func nodeAt(index: Int) -> Node<T>? {            // get the node at the specified index
        if index >= 0 {                                     // make sure the index is positive
            var node = head                                 // start at the head of the list
            var i = index                                   // the count down
            while node != nil {                             // iterate over the list
                if i == 0 { return node }                   // until we finished the countdown and return the node
                i -= 1                                      // countdown
                node = node!.next                           // and take the next item
            }
        }
        return nil                                          // index is out of range
    }
    
    public subscript(index: Int) -> T {                     // allow the list to be subscripted
        let node = nodeAt(index: index)                     // get the node at the specified index
        assert(node != nil)                                 // make sure, its not nil
        return node!.value                                  // and return its value
    }
    
    public func indexOf(node n: Node<T>) -> Int? {          // get the index of a supplied node in the list
        var i = 0                                           // start at the beginning
        var node = head                                     // with the head of the list
        while node != nil {                                 // and iterate over the whole list
            if node! === n { return i }                     // until we find the identical node: return its index
            i += 1                                          // increase the index
            node = node!.next                               // and take the next item
        }
        return nil                                          // could not find the supplied node
    }

    public func removeAll() {                               // empty the list
        head = nil                                          // clear its head
        tail = nil                                          // and its tail
        count = 0                                           // and reset the counter
    }
    
    public func remove(node: Node<T>) -> T {                // remove a node from the list
        let prev = node.previous                            // first get its neighbours
        let next = node.next
        
        if let prev = prev {                                // we are not at the top of the list
            prev.next = next                                // make the previous node point to our following one
        } else {                                            // we are at the top of the list
            head = next                                     // make our follower the new head of the list
        }
        next?.previous = prev                               // and make our follower point to our previous node
        
        if next == nil {                                    // we are at the end of the list and do not have a follower
            tail = prev                                     // make the previous node the new tail of the list
        }
        
        node.previous = nil                                 // and clear the pointers in our node
        node.next = nil
        count -= 1                                          // decrease the counter
        
        return node.value                                   // and return our object
    }
    
    private func nodesBeforeAndAfter(index: Int) -> (NodeType?, NodeType?) {    // get the nodes before and after an index
        assert(index >= 0)                                  // make sure the index is positive (+ 0)
        
        var i = index                                       // let's countdown
        var next = head                                     // start at the top of the list
        var prev: NodeType?                                 // in the beginning the previous node is nil
        
        while next != nil && i > 0 {                        // iterate over the list
            i -= 1                                          // and count down
            prev = next                                     // the previous node becomes the current one
            next = next!.next                               // and the current one the next
        }
        assert(i == 0)                                      // make sure we hit the specified index
        
        return (prev, next)                                 // and return the previous and the current node
    }
    
    private func insert(value: T, prev: NodeType?, next: NodeType?) {   // insert a new node between neighbours
        let newNode = Node(value: value, list: self)        // create the new node from the object
        newNode.previous = prev                             // let's point to the old neighbours
        newNode.next = next
        prev?.next = newNode                                // and from the neighbours to our new node
        next?.previous = newNode
        
        if prev == nil {                                    // we inserted at the top of the list
            head = newNode                                  // the head will be our new node
        }
        if next == nil {                                    // we inserted at the end of the list
            tail = newNode                                  // the tail will be our new node
        }
        count += 1                                          // and increase the counter
    }
    
    public func insert(value: T, atIndex index: Int) {      // insert an object at a specified index
        let (prev, next) = nodesBeforeAndAfter(index: index)    // get the neighbours at the index
        insert(value: value, prev: prev, next: next)        // and insert it in between
    }

    public func insert(value: T, afterNode node: Node<T>?) {    // insert an object after a supplied node
        if node == nil {                                    // we want to insert at the top of the list
            insert(value: value, atIndex: 0)                // make the new node be index 0
        } else {                                            // not at the top of the list
            insert(value: value, prev: node, next: node!.next)  // insert it between the supplied node and its follower
        }
    }
    
    public func insert(value: T, beforeNode node: Node<T>?) {   // insert an object before a supplied node
        if node == nil {                                    // we want to insert at the end of the list
            append(value: value)                            // just append the object
        } else {                                            // not at the end of the list
            insert(value: value, prev: node!.previous, next: node)  // insert it between the previous node and its follower
        }
    }
}

typealias PropertyList = [String : Any]

protocol PropertyListReadable {
    func propertyListRepresentation() -> PropertyList
    init?(propertyListRepresentation:PropertyList?)
}

extension LinkedList: Codable where T: Codable {
    public convenience init(from decoder: Decoder) throws {
        var values = try decoder.unkeyedContainer()         // get the container from the decoder
        self.init()
        if let list = try? values.decode([T].self) {        // and decode the array
            for object in list {                            // iterate over all elements
                self.append(value: object)                  // and append it to our list
            }
        } else {
            throw CodingError.decoding("Decoding Error: \(dump(values))")   // had no success
        }
    }
    
    public func encode(to encoder: Encoder) throws {
        var container = encoder.unkeyedContainer()          // get an empty container first
        var next = head                                     // start at the top
        var list = [T]()                                    // create an array of the objects
        while next != nil {                                 // iterate over the whole list
            list.append(next!.value)                        // and append our objects
            next = next!.next                               // get the next item
        }
        try container.encode(list)                          // and encode the list
    }
}

struct PersonPLR: PropertyListReadable {
    var name: String
    var age: Int
    
    init(name: String, age: Int) {
        self.name = name
        self.age = age
    }
    
    init?(propertyListRepresentation: PropertyList?) {
        guard let values = propertyListRepresentation else {return nil}
        if let name = values["name"] as? String,
            let age = values["age"] as? Int {
            self.name = name
            self.age = age
        } else {
            return nil
        }
    }
    
    func propertyListRepresentation() -> PropertyList {
        let representation: PropertyList = ["name" : self.name as Any,
                                            "age" : self.age as Any]
        return representation
    }
}

let mePLR = PersonPLR(name: "Frank-Peter", age: 62)
let personArrayPLR = [mePLR, mePLR]
let personListPLR = LinkedList<PersonPLR>(personArrayPLR)
print(personListPLR)

let dataPLR = try NSKeyedArchiver.archivedData(withRootObject: personListPLR, requiringSecureCoding: false)
let personListPLR2 = try NSKeyedUnarchiver.unarchiveTopLevelObjectWithData(dataPLR) as! LinkedList<PersonPLR>
print(personListPLR2)

class PersonCodable: Codable, CustomStringConvertible {
    var name: String
    var age: Int

    init(name: String, age: Int) {
        self.name = name
        self.age = age
    }
    public var description: String {
        let text = "PersonCodable(name: \"\(name)\", age: \(age))"
        return text
    }
}

let meCodable = PersonCodable(name: "Frank-Peter", age: 62)
let personArrayCodable = [meCodable, meCodable]
let personListCodable = LinkedList<PersonCodable>(personArrayCodable)
print(personListCodable)

var identical = personListCodable[0] === personListCodable[1] ? "===" : "!==="
print("[0] \(identical) [1]")

let dataCodable = try PropertyListEncoder().encode(personListCodable)
let personListCodable2 = try PropertyListDecoder().decode(LinkedList<PersonCodable>.self, from: dataCodable)
print(personListCodable2)

identical = personListCodable2[0] === personListCodable2[1] ? "===" : "!==="
print("[0] \(identical) [1]")

class PersonNSCoding: NSObject, NSCoding {
    var name: String = ""
    var age: Int = 0
    
    init(name: String, age: Int) {
        super.init()
        self.name = name
        self.age = age
    }
    
    required init?(coder aDecoder: NSCoder) {
        super.init()
        self.name = aDecoder.decodeObject(forKey: "name") as! String
        self.age = aDecoder.decodeInteger(forKey: "age")
    }
    
    func encode(with aCoder: NSCoder) {
        aCoder.encode(self.name, forKey: "name")
        aCoder.encode(self.age, forKey: "age")
    }
    
    override public var description: String {
        let text = "PersonNSCoding(name: \"\(name)\", age: \(age))"
        return text
    }
}

let meNSCoding = PersonNSCoding(name: "Frank-Peter", age: 62)
let personArrayNSCoding = [meNSCoding, meNSCoding]
let personListNSCoding = LinkedList<PersonNSCoding>(personArrayNSCoding)
print(personListNSCoding)

let dataNSCoding = try NSKeyedArchiver.archivedData(withRootObject: personListNSCoding, requiringSecureCoding: false)
let personListNSCoding2 = try NSKeyedUnarchiver.unarchiveTopLevelObjectWithData(dataNSCoding) as! LinkedList<PersonNSCoding>
print(personListNSCoding2)

identical = personListNSCoding2[0] === personListNSCoding2[1] ? "===" : "!==="
print("[0] \(identical) [1]")
